"""Aircrack-ng."""
from .executor import ExecutorHelper


class AircrackNg(ExecutorHelper):
    """Aircrack-ng 1.6  - (C) 2006-2020 Thomas d'Otreppe
    https://www.aircrack-ng.org

    Usage: aircrack-ng [options] <file>...

    Options:

        -a <amode> : force attack mode (1/WEP, 2/WPA-PSK)
        -e <essid> : target selection: network identifier
        -b <bssid> : target selection: access point's MAC
        -p <nbcpu> : # of CPU to use  (default: all CPUs)
        -q         : enable quiet mode (no status output)
        -C <macs>  : merge the given APs to a virtual one
        -l <file>  : write key to file. Overwrites file.
        -c         : search alpha-numeric characters only
        -t         : search binary coded decimal chr only
        -h         : search the numeric key for Fritz!BOX
        -d <mask>  : use masking of the key (A1:XX:CF:YY)
        -m <maddr> : MAC address to filter usable packets
        -n <nbits> : WEP key length :  64/128/152/256/512
        -i <index> : WEP key index (1 to 4), default: any
        -f <fudge> : bruteforce fudge factor,  default: 2
        -k <korek> : disable one attack method  (1 to 17)
        -x or -x0  : disable bruteforce for last keybytes
        -x1        : last keybyte bruteforcing  (default)
        -x2        : enable last  2 keybytes bruteforcing
        -X         : disable  bruteforce   multithreading
        -y         : experimental  single bruteforce mode
        -K         : use only old KoreK attacks (pre-PTW)
        -s         : show the key in ASCII while cracking
        -M <num>   : specify maximum number of IVs to use
        -D         : WEP decloak, skips broken keystreams
        -P <num>   : PTW debug:  1: disable Klein, 2: PTW
        -1         : run only 1 try to crack key with PTW
        -V         : run in visual inspection mode
        -w <words> : path to wordlist(s) filename(s)
        -N <file>  : path to new session filename
        -R <file>  : path to existing session filename
        -E <file>  : create EWSA Project file v3
        -I <str>   : PMKID string (hashcat -m 16800)
        -j <file>  : create Hashcat v3.6+ file (HCCAPX)
        -J <file>  : create Hashcat file (HCCAP)
        -S         : WPA cracking speed test
        -Z <sec>   : WPA cracking speed test length of
                     execution.
        -r <DB>    : path to airolib-ng database
        --simd-list       : Show a list of the available
        --simd <option>   : Use specific SIMD architecture.
        -u         : Displays # of CPUs & SIMD support
        --help     : Displays this usage screen
    """

    command = "aircrack-ng"
    requires_tempfile = True
    requires_tempdir = False
    requires_root = False

    async def run(self, *args, **kwargs):
        if self.tempfile:
            kwargs["l"] = self.tempfile.name
        return await super().run(*args, **kwargs)

    async def get_result(self):
        """Read tempfile (write key result)"""
        if self.proc:
            await self.proc.wait()
        if self.tempfile:
            return self.tempfile.read()



from behave import then, when


@when("I import the package pyrcrack")
def import_package(context):
    """Try to import the package.

    If an exception happened push to context.
    """
    context.exceptions = []
    try:
        import pyrcrack  # noqa
    except ImportError as excp:
        context.exceptions = [excp]


@then("I see no errors")
def no_exceptions_in_context(context):
    """Check that no previous exceptions have been captured."""
    assert not context.exceptions



"""Scan for targets and and pretty print some data."""
import asyncio
import logging

from rich.console import Console
from rich.prompt import Prompt

import pyrcrack

logging.basicConfig(level=logging.DEBUG)


async def scan_for_targets():
    """Scan for targets, return json."""
    console = Console()
    console.clear()
    console.show_cursor(show=False)
    airmon = pyrcrack.AirmonNg()
    interfaces = await airmon.interfaces
    console.print(interfaces.table)
    interface = Prompt.ask("Select an interface",
                           choices=[a.interface for a in interfaces])

    async with airmon(interface) as mon:
        async with pyrcrack.AirodumpNg() as pdump:
            async for aps in pdump(mon.monitor_interface):
                console.clear()
                console.print(aps.table)
                client = Prompt.ask(
                    "Select an AP",
                    choices=["continue", *[str(a) for a in range(len(aps))]],
                )

                if client != "continue":
                    break

        async with pyrcrack.AirodumpNg() as pdump:
            console.print(
                ":vampire:",
                f"Selected client: [red] {aps[int(client)].bssid} [/red]")

            async for result in pdump(mon.monitor_interface,
                                      **aps[int(client)].airodump):
                console.clear()
                console.print(result.table)
                await asyncio.sleep(3)


asyncio.run(scan_for_targets())



"""Airdecap-ng."""
from .executor import ExecutorHelper


class AirdecapNg(ExecutorHelper):
    """Airdecap-ng 1.6  - (C) 2006-2020 Thomas d'Otreppe

    https://www.aircrack-ng.org

    Usage: airdecap-ng [options] <pcap file>

    Options:

        -l         : don't remove the 802.11 header
        -b <bssid> : access point MAC address filter
        -e <essid> : target network SSID
        -o <fname> : output file for decrypted packets (default <src>-dec)
        -w <key>   : target network WEP key in hex
        -c <fname> : output file for corrupted WEP packets (default <src>-bad)
        -p <pass>  : target network WPA passphrase
        -k <pmk>   : WPA Pairwise Master Key in hex
        --help     : Displays this usage screen

    """

    command = "airdecap-ng"
    sync = False
    requires_tempfile = False
    requires_tempdir = False
    requires_root = False



"""Aireplay-ng."""
import asyncio

from .executor import ExecutorHelper
from .models import AireplayResults


class AireplayNg(ExecutorHelper):
    """Aireplay-ng 1.6  - (C) 2006-2020 Thomas d'Otreppe
    https://www.aircrack-ng.org

    Usage: aireplay-ng <options> <replay interface>

    Options:

        -b bssid  : MAC address, Access Point
        -d dmac   : MAC address, Destination
        -s smac   : MAC address, Source
        -m len    : minimum packet length
        -n len    : maximum packet length
        -u type   : frame control, type    field
        -v subt   : frame control, subtype field
        -t tods   : frame control, To      DS bit
        -f fromds : frame control, From    DS bit
        -w iswep  : frame control, WEP     bit
        -D        : disable AP detection
        -x nbpps  : number of packets per second
        -p fctrl  : set frame control word (hex)
        -a bssid  : set Access Point MAC address
        -c dmac   : set Destination  MAC address
        -h smac   : set Source       MAC address
        -g value  : change ring buffer size (default: 8)
        -F        : choose first matching packet
        -e essid  : set target AP SSID
        -o npckts : number of packets per burst (0=auto, default: 1)
        -q sec    : seconds between keep-alives
        -Q        : send reassociation requests
        -y prga   : keystream for shared key auth
        -T n      : exit after retry fake auth request n time
        -j        : inject FromDS packets
        -k IP     : set destination IP in fragments
        -l IP     : set source IP in fragments
        -B        : activates the bitrate test
        -i iface  : capture packets from this interface
        -r file   : extract packets from this pcap file
        -R                    : disable /dev/rtc usage
        --ignore-negative-one  : if the interface's channel can't be determined
        --deauth-rc <rc>      : Deauthentication reason code [0-254]
        --deauth <count>    : deauthenticate 1 or all stations (-0)
        --fakeauth <delay>  : fake authentication with AP (-1)
        --interactive       : interactive frame selection (-2)
        --arpreplay         : standard ARP-request replay (-3)
        --chopchop          : decrypt/chopchop WEP packet (-4)
        --fragment          : generates valid keystream   (-5)
        --caffe-latte       : query a client for new IVs  (-6)
        --cfrag             : fragments against a client  (-7)
        --migmode           : attacks WPA migration mode  (-8)
        --test              : tests injection and quality (-9)
        --help              : Displays this usage screen
    """

    command = "aireplay-ng"
    requires_tempfile = False
    requires_tempdir = False
    requires_root = True

    async def run(self, *args, **kwargs):
        """Run async, with prefix stablished as tempdir."""
        asyncio.create_task(self.result_updater())
        return await super().run(*args, **kwargs)

    @property
    async def results(self):
        return self.meta.get("result", AireplayResults(""))

    async def result_updater(self):
        """Set result on local object."""
        # TODO: There might be a situation here where proc is never != None.
        # This should have some timeout logic as airodump has
        while not self.proc:
            await asyncio.sleep(1)

        while self.proc.returncode is None:
            self.meta["result"] = await self.get_results()
            await asyncio.sleep(2)

    async def get_results(self):
        """Get results list."""
        if not self.proc:
            return AireplayResults("")
        return AireplayResults((await self.proc.communicate())[0].decode())



"""Airmon-ng."""
import asyncio
import re

from .executor import ExecutorHelper
from .models import Interfaces


class AirmonNg(ExecutorHelper):
    """Airmon-NG
    Usage: airmon-ng <start|stop|check> <interface> [channel or frequency]
    """

    command = "airmon-ng"
    requires_tempfile = False
    requires_tempdir = False
    requires_root = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dirty = False
        self.monitor_token = None
        self.monitor_enabled = []

    async def run(self, *args, **kwargs):
        """Check argument position.

        Forced for this one.
        """
        self.dirty = True
        if args:
            if not any(a in args[0] for a in ("start", "stop", "check")):
                raise Exception("First argument for airmon must"
                                f" be start, stop or check not {args[0]}")

        return await super().run(*args, **kwargs)

    async def __aenter__(self):
        """Put selected interface in monitor mode."""
        if not self.run_args[0][0]:
            raise RuntimeError("Should be called (airmon()) first.")
        ifaces = await self.interfaces
        if not any(a.interface == self.run_args[0][0] for a in ifaces):
            if not any(a == self.run_args[0][0] for a in ifaces):
                raise ValueError("Invalid interface selected")
        await self.run("start", self.run_args[0][0])
        # Save interface data while we're on the async cm.
        self._interface_data = await self.interfaces
        from . import MONITOR

        self.monitor_token = MONITOR.set(self.monitor_interface)
        return self

    async def select_interface(self, regex):
        ifaces = await self.interfaces
        reg = re.compile(regex)
        if interface := next((a for a in ifaces if reg.match(str(a))), None):
            return interface
        raise Exception(f"No interface matching regex {regex}")

    async def __aexit__(self, *args, **kwargs):
        """Set monitor-enabled interfaces back to normal."""
        await self.run("stop", self.monitor_interface)
        self.dirty = False
        await asyncio.sleep(1)
        if self.monitor_token:
            from . import MONITOR

            MONITOR.reset(self.monitor_token)
        await self.interfaces

    @property
    def monitor_interface(self):
        iface = next(a for a in self._interface_data
                     if a.interface == self.run_args[0][0])
        return iface.monitor

    @property
    async def interfaces(self):
        """List of currently available interfaces as reported by airmon-ng.

        This is an awaitable property, use it as in::

        async with AirmonNg() as airmon:
            await airmon.interfaces

        Returns: None
        """
        if not self.dirty:
            await self.run()
            self.dirty = False
        data = await self.readlines()
        return Interfaces(data)

    def __str__(self):
        return self.monitor_interface



import pytest


@pytest.mark.asyncio
async def test_run_airmon():
    """Test main run method."""
    import asynctest

    from pyrcrack import AirmonNg

    with asynctest.mock.patch("asyncio.create_subprocess_exec") as runmock:
        with pytest.raises(Exception):
            await AirmonNg().run("foo")
        await AirmonNg().run("start")

        runmock.assert_called_with(
            "airmon-ng", "start", stderr=-1, stdout=-1, stdin=-1)

    with asynctest.mock.patch("asyncio.create_subprocess_exec") as runmock:
        await AirmonNg().run()
        runmock.assert_called_with("airmon-ng", stderr=-1, stdout=-1, stdin=-1)


def test_run_model():
    from pyrcrack.models import Interfaces

    data = [
        b"Found 6 processes that could cause trouble.",
        b"Kill them using 'airmon-ng check kill' before putting",
        b"the card in monitor mode, they will interfere by changing channels",
        b"and sometimes putting the interface back in managed mode",
        b"  PID Name",
        b"  239 avahi-daemon",
        b"  241 wpa_supplicant",
        b"  243 avahi-daemon",
        b"  356 wpa_supplicant",
        b"  381 wpa_supplicant",
        b"  481 dhcpcd",
        b"PHY\tInterface\tDriver\t\tChipset",
        b"phy0\twlan0\t\tbrcmfmac\tBroadcom 43430",
        (
            b"\t\t(mac80211 monitor mode vif enabled for [phy0]wlan0"
            b" on [phy0]wlan0mon)"
        ),
        b"\t\t(mac80211 station mode vif disabled for [phy0]wlan0)",
        b"phy1\twlxe84e066b5386\tmt7601u\t\tRalink Technology, Corp. MT7601U",
    ]
    assert [a.__dict__ for a in Interfaces(data)] == [
        {
            "data": {
                "phy": "phy0",
                "interface": "wlan0",
                "driver": "brcmfmac",
                "chipset": "Broadcom 43430",
                "monitor": {
                    "driver": "mac80211",
                    "interface": "wlan0mon",
                    "mode": "monitor",
                    "original_interface": "wlan0",
                    "status": "enabled",
                },
            }
        },
        {
            "data": {
                "phy": "phy1",
                "interface": "wlxe84e066b5386",
                "driver": "mt7601u",
                "chipset": "Ralink Technology, Corp. MT7601U",
            }
        },
    ]

    data = b"""Found 7 processes that could cause trouble.
Kill them using 'airmon-ng check kill' before putting
the card in monitor mode, they will interfere by changing channels
and sometimes putting the interface back in managed mode

    PID Name
    677 wpa_supplicant
   2209 dhcpcd
   2210 dhcpcd
   2211 dhcpcd
 107797 dhcpcd
 107805 dhcpcd
 107853 dhcpcd

PHY	Interface	Driver		Chipset

phy0	wlp3s0		iwlwifi		Intel Corporation Wireless 7260 (rev 83)

\t\t(mac80211 monitor mode vif enabled for [phy0]wlp3s0 on [phy0]wlp3s0mon)
\t\t(mac80211 station mode vif disabled for [phy0]wlp3s0)
""".split(
        b"\n"
    )

    assert [a.__dict__ for a in Interfaces(data)] == [
        {
            "data": {
                "phy": "phy0",
                "interface": "wlp3s0",
                "driver": "iwlwifi",
                "chipset": "Intel Corporation Wireless 7260 (rev 83)",
                "monitor": {
                    "driver": "mac80211",
                    "interface": "wlp3s0mon",
                    "mode": "monitor",
                    "original_interface": "wlp3s0",
                    "status": "enabled",
                },
            }
        }
    ]



"""Test executor class."""

import pytest


def test_helpstr_extraction():
    """Test helpstr extraction.

    This tests helpstr(), wich should try to execute a command, and echo
    to ensure output 0.
    """
    from unittest.mock import patch

    with patch("subprocess.check_output", side_effect=(b"test",)) as mock:
        from pyrcrack.executor import ExecutorHelper

        class FakeExecutor(ExecutorHelper):
            command = "foobar"
            requires_tempfile = False
            requires_tempdir = False
            requires_root = False

        assert FakeExecutor().helpstr == "test"
        mock.assert_called_with("foobar 2>&1; echo", shell=True)


def test_usage():
    """Test extract usage from a specified command."""
    # opt = docopt.parse_defaults(self.helpstr)
    # return dict({a.short or a.long: bool(a.argcount) for a in opt})
    from unittest.mock import patch

    with patch("subprocess.check_output", side_effect=(b"test",)):
        from pyrcrack.executor import ExecutorHelper

        class FakeExecutor(ExecutorHelper):
            """Fake command.
            Usage: fake [options]

            Options:
                -f
                -y=<foo>
            """

            command = "foobar"
            requires_tempfile = False
            requires_tempdir = False
            requires_root = False

        assert FakeExecutor().usage == {"-f": False, "-y": True}


@pytest.mark.asyncio
async def test_run_async():
    """Check command usage."""
    import asynctest

    with asynctest.mock.patch("asyncio.create_subprocess_exec") as runmock:
        from pyrcrack.executor import ExecutorHelper

        class FakeExecutor(ExecutorHelper):
            """Fake command.
            Usage: fake [options]

            Options:
                -f=<bar>
                -y
            """

            command = "foobar"
            requires_tempfile = False
            requires_tempdir = False
            requires_root = False

        await FakeExecutor().run(f="foo", y=True)

        try:
            runmock.assert_called_with(
                *["foobar", "-f", "foo", "-y"], stderr=-1, stdin=-1, stdout=-1
            )
        except AssertionError:
            runmock.assert_called_with(
                *["foobar", "-y", "-f", "foo"], stderr=-1, stdin=-1, stdout=-1
            )


@pytest.mark.parametrize(
    "in_,out", (("aircrack-ng", "AircrackNg"), ("airodump-ng", "AirodumpNg"))
)
def test_stc(in_, out):
    """Convert snake case to camelcase in class format."""
    from pyrcrack.executor import stc

    assert stc(in_) == out



"""Pyrcrack-ng Executor helper."""
import asyncio
import functools
import itertools
import logging
import os
import subprocess
import tempfile
import uuid
import warnings
from contextlib import suppress
from contextvars import ContextVar

import docopt
import stringcase

logging.basicConfig(level=logging.INFO)


class Option:
    """Represents a single option (e.g, -e)."""

    def __init__(self, usage, word=None, value=None, logger=None):
        """Set option parameters."""
        self.usage = usage
        self.word = word
        self.logger = logger
        self.value = value
        keys = usage.keys()
        self.is_short = Option.short(word) in keys
        self.is_long = Option.long(word) in keys
        self.expects_args = bool(usage[self.formatted])
        self.logger.debug("Parsing option %s:%s", self.word, self.value)

    @property
    @functools.lru_cache  # noqa: B019
    def formatted(self):
        """Format given option acording to definition."""
        result = Option.short(self.word) if self.is_short else Option.long(
            self.word)

        if self.usage.get(result):
            return result

        sword = self.word.replace("_", "-")
        return Option.short(sword) if self.is_short else Option.long(sword)

    @staticmethod
    def long(word):
        """Extract long format option."""
        return "--{}".format(word)

    @staticmethod
    def short(word):
        """Extract short format option."""
        return "-{}".format(word)

    @property
    def parsed(self):
        """Returns key, value if value is required."""
        if self.expects_args:
            return (self.formatted, str(self.value))
        return (self.formatted, )

    def __repr__(self):
        return f"Option(<{self.parsed}>, {self.is_short}, {self.expects_args})"


def check():
    """Check if aircrack-ng is compatible."""
    ver_check = subprocess.check_output(["aircrack-ng", "--help"])
    if b"Aircrack-ng 1.6" not in ver_check:
        if b"Aircrack-ng 1.7" not in ver_check:
            raise Exception("Unsupported Aircrack-ng detected")
        else:
            warnings.warn(
                "Aircrack-ng 1.7 detected, some features may not work")


class ExecutorHelper:
    """Abstract class interface to a shell command."""

    requires_tempfile = False
    requires_tempdir = False
    requires_root = True
    command = ""

    def __init__(self):
        """Set docstring."""
        if not self.__doc__:
            self.__doc__ = self.helpstr
        self.uuid = uuid.uuid4().hex
        self.called = False
        self.execn = 0
        self.logger = logging.getLogger(self.__class__.__name__)
        self.proc = None
        self.meta = {}
        self.debug = os.getenv("PYRCRACK_DEBUG", "") == 1
        self.tempfile = None
        self.tempdir = None
        if self.requires_tempfile:
            self.tempfile = tempfile.NamedTemporaryFile()
        elif self.requires_tempdir:
            self.tempdir = tempfile.TemporaryDirectory()
        if self.requires_root and not os.getenv("SKIP_ROOT_CHECK"):
            if os.geteuid() != 0:
                raise Exception("Must be run as root")
        if not os.getenv("SKIP_VERSION_CHECK"):
            check()

    @property
    @functools.lru_cache  # noqa: B019
    def helpstr(self):
        """Extract help string for current command."""
        helpcmd = "{} 2>&1; echo".format(self.command)
        return subprocess.check_output(helpcmd, shell=True).decode()

    @property
    @functools.lru_cache  # noqa: B019
    def usage(self):
        """Extract usage from a specified command.

        This is useful if usage not defined in subclass, but it is
        recommended to define them there.
        """
        opt = docopt.parse_defaults(self.__doc__)
        return dict({a.short or a.long: bool(a.argcount) for a in opt})

    def _run(self, *args, **kwargs):
        """Check command usage and execute it.

        If self.sync is defined, it will return process call output,
        and launch it blockingly.

        Otherwise it will call asyncio.create_subprocess_exec()
        """
        if not self.command:
            raise Exception("Subclassing error, please specify a base cmd")

        self.logger.debug("Parsing options: %s", kwargs)
        options = [
            Option(self.usage, a, v, self.logger) for a, v in kwargs.items()
        ]
        self.logger.debug("Got options: %s", options)

        opts = ([self.command] + list(args) +
                list(itertools.chain(*(o.parsed for o in options))))

        self.logger.debug("Running command: %s", opts)
        return opts

    @staticmethod
    def resolve(val):
        """Force string conversion.

        - In case an Interface object comes on args
        - In case a contextvar comes, retrieve its value
        """
        # I'm a little conflicted on this one, as it seems dirty to cast
        # everything to string
        # Yet, I'm letting it go like this because in the end, this is a
        # subprocess call that requires all args to be strings or bytes...
        # This will also allow us to use an AP object in a future.
        if isinstance(val, ContextVar):
            return str(val.get())
        return str(val)

    async def run(self, *args, **kwargs):
        """Run asynchronously."""
        opts = self._run(*args, **kwargs)
        self.proc = await asyncio.create_subprocess_exec(
            *[self.resolve(a) for a in opts],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return self.proc

    def __call__(self, *args, **kwargs):
        self.run_args = args, kwargs
        return self

    def __aiter__(self):
        """Defines us as an async iterator."""
        return self

    async def __anext__(self):
        """Get the next result batch."""
        if not self.called:
            self.called = True
            self.proc = await self.run(*self.run_args[0], **self.run_args[1])

        if not asyncio.iscoroutine(self.results):
            raise RuntimeError('Results must be a coroutine')

        current_results = await self.results

        if not self.running:
            # If we haven't sent the latests results, send them now.
            if self.last_results != current_results:
                self.last_results = current_results
                return current_results
            raise StopAsyncIteration

        self.last_results = current_results
        return self.last_results

    @property
    def running(self):
        if not self.proc:
            return False
        return self.proc.returncode is None

    async def readlines(self):
        """Return lines as per proc.communicate, non-empty ones."""
        if not self.proc:
            return []
        com = await self.proc.communicate()
        return [a for a in com[0].split(b"\n") if a]

    @property
    async def results(self):
        return [self.proc]

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Clean up conext manager."""
        if self.debug and (self.requires_tempfile or self.requires_tempdir):
            self.logger.error(f"Not deleting {self.tempfile}, {self.tempdir}")

        if self.tempfile:
            self.tempfile.__exit__(exc_type, exc_val, exc_tb)

        if self.tempdir:
            self.tempdir.__exit__(exc_type, exc_val, exc_tb)

        if exc_type in (ProcessLookupError, subprocess.CalledProcessError):
            return True

        if self.proc:
            with suppress(Exception):
                self.proc.kill()

    async def __aenter__(self):
        """Create temporary directories and files if required."""
        if self.requires_tempfile and self.tempfile:
            self.tempfile.__enter__()
        elif self.requires_tempdir and self.tempdir:
            self.tempdir.__enter__()
        return self


def stc(command):
    """Convert snake case to camelcase in class format."""
    return stringcase.pascalcase(command.replace("-", "_"))



"""Deauth"""
import asyncio
from contextlib import suppress

from rich.console import Console

import pyrcrack

CONSOLE = Console()
CONSOLE.clear()
CONSOLE.show_cursor(show=False)


async def attack(apo):
    """Run aireplay deauth attack."""
    airmon = pyrcrack.AirmonNg()
    interfaces = await airmon.interfaces
    CONSOLE.print("Starting airmon-ng with channel {}".format(apo.channel))
    interface = interfaces[0]
    async with airmon(interface.interface, apo.channel) as mon:
        async with pyrcrack.AireplayNg() as aireplay:
            CONSOLE.print("Starting aireplay on {} for {}".format(
                mon.monitor_interface, apo.bssid))
            await aireplay.run(mon.monitor_interface,
                               deauth=10,
                               D=True,
                               b=apo.bssid)
            while True:
                CONSOLE.print(aireplay.meta)
                await asyncio.sleep(2)


async def deauth():
    """Scan for targets, return json."""
    airmon = pyrcrack.AirmonNg()
    interfaces = await airmon.interfaces
    async with airmon(interfaces[0].interface) as mon:
        async with pyrcrack.AirodumpNg() as pdump:
            async for result in pdump(mon.monitor_interface):
                CONSOLE.print(result.table)
                with suppress(KeyError):
                    ap = result[0]
                    CONSOLE.print("Selected AP {}".format(ap.bssid))
                    break
                await asyncio.sleep(3)
    await attack(ap)


asyncio.run(deauth())



"""pyrcrack.

Aircrack-NG python bindings
"""

from contextvars import ContextVar

from .airbase import AirbaseNg  # noqa
from .aircrack import AircrackNg  # noqa
from .airdecap import AirdecapNg  # noqa
from .airdecloack import AirdecloackNg  # noqa
from .aireplay import AireplayNg  # noqa
from .airmon import AirmonNg  # noqa
from .airodump import AirodumpNg  # noqa

MONITOR = ContextVar("monitor_interface")

__version__ = "1.2.6"



"""Test specific airodump functions."""
import pytest


def test_props():
    """Props."""
    from pyrcrack import AirodumpNg

    assert AirodumpNg.requires_tempdir
    assert not AirodumpNg.requires_tempfile
    assert AirodumpNg.command == "airodump-ng"


@pytest.mark.asyncio
async def test_run_airodump():
    """Test main run method."""
    import asynctest

    from pyrcrack import AirodumpNg

    async with AirodumpNg() as airodump:
        with asynctest.mock.patch("asyncio.create_subprocess_exec") as runmock:
            airodump.result_updater = asynctest.mock.CoroutineMock()
            await airodump.run()
            assert runmock.called
            runmock.assert_called_with(
                "airodump-ng",
                "--background",
                "1",
                "--write",
                airodump.tempdir.name + "/" + airodump.uuid,
                "--write-interval",
                "1",
                "--output-format",
                "netxml,logcsv,netxml",
                stderr=-1,
                stdout=-1,
                stdin=-1,
            )


@pytest.mark.asyncio
async def test_run_airodump_with_write():
    """Test main run method."""
    import asynctest

    from pyrcrack import AirodumpNg

    async with AirodumpNg() as airodump:
        with asynctest.mock.patch("asyncio.create_subprocess_exec") as runmock:
            await airodump.run(write="foo")
            runmock.assert_called_with(
                "airodump-ng",
                "--background",
                "1",
                "--write",
                "foo",
                "--write-interval",
                "1",
                "--output-format",
                "netxml,logcsv,netxml",
                stderr=-1,
                stdout=-1,
                stdin=-1,
            )


@pytest.mark.asyncio
async def test_csv():
    """Test main run method."""
    from unittest.mock import MagicMock

    from pyrcrack import AirodumpNg

    async with AirodumpNg() as airodump:
        expected = (
            f"{airodump.tempdir.name}/{airodump.uuid}"
            f"-{airodump.execn:02}.kismet.csv"
        )
        airodump.proc = MagicMock()
        assert airodump.get_file("kismet.csv") == expected



"""Airdecloack-ng."""

from .executor import ExecutorHelper


class AirdecloackNg(ExecutorHelper):
    """Airdecloak-ng 1.6  - (C) 2008-2020 Thomas d'Otreppe
    https://www.aircrack-ng.org
    Available filters::

         signal:               Try to filter based on signal.
         duplicate_sn:         Remove all duplicate sequence numbers
                               for both the AP and the client.
         duplicate_sn_ap:      Remove duplicate sequence number for
                               the AP only.
         duplicate_sn_client:  Remove duplicate sequence number for the
                               client only.
         consecutive_sn:       Filter based on the fact that IV should
                               be consecutive (only for AP).
         duplicate_iv:         Remove all duplicate IV.
         signal_dup_consec_sn: Use signal (if available), duplicate and
                               consecutive sequence number (filtering is
                                much more precise than using all these
                                filters one by one).

    Usage: airdecloak-ng [options]

    Options:
       -i <file>             : Input capture file
       --ssid <ESSID>        : ESSID of the network to filter
       --bssid <BSSID>       : BSSID of the network to filter
       -o <file>             : Output packets (valid) file
       -c <file>             : Output packets (cloaked) file
       -u <file>             : Output packets (unknown/ignored) file
       --filters <filters>   : Apply filters (separated by a comma).
       --null-packets        : Assume that null packets can be cloaked.
       --disable-base_filter : Do not apply base filter.
       --drop-frag           : Drop fragmented packets
       --help                : Displays this usage screen
    """

    command = "airdecloack-ng"
    requires_tempfile = False
    requires_tempdir = False
    requires_root = False



"""Test specific aireplay functions."""
import pytest


def test_aireplay_props():
    """Props."""
    from pyrcrack import AireplayNg

    assert not AireplayNg.requires_tempdir
    assert not AireplayNg.requires_tempfile
    assert AireplayNg.command == "aireplay-ng"


@pytest.mark.asyncio
async def test_run_aireplay():
    """Test main run method."""
    import asynctest

    from pyrcrack import AireplayNg

    async with AireplayNg() as aireplay:
        with asynctest.mock.patch("asyncio.create_subprocess_exec") as runmock:
            aireplay.result_updater = asynctest.mock.CoroutineMock()
            await aireplay.run()
            assert runmock.called
            runmock.assert_called_with(
                "aireplay-ng", stderr=-1, stdout=-1, stdin=-1)



"""Airbase-ng."""

from .executor import ExecutorHelper


class AirbaseNg(ExecutorHelper):
    """Airbase-ng 1.6  - (C) 2008-2020 Thomas d'Otreppe
    Original work: Martin Beck
    https://www.aircrack-ng.org

    Usage: airbase-ng <options> <replay interface>

    Options:

        -a <bssid>       : set Access Point MAC address
        -i <iface>       : capture packets from this interface
        -w <key>         : use this WEP key to en-/decrypt packets
        -h <MAC>         : source mac for MITM mode
        -f <disallow>    : disallow specified client MACs (default: allow)
        -W 0|1           : [don't] set WEP flag in beacons 0|1 (default: auto)
        -q               : quiet (do not print statistics)
        -v               : verbose (print more messages)
        -A               : Ad-Hoc Mode (allows other clients to peer)
        -Y <proc>        : (in|out|both) external packet processing
        -c <channel>     : sets the channel the AP is running on
        -X               : hidden ESSID
        -s               : force shared key authentication (default: auto)
        -S               : set shared key challenge length (default: 128)
        -L               : Caffe-Latte WEP attack
        -N               : cfrag WEP attack (recommended)
        -x <nbpps>       : number of packets per second (default: 100)
        -y               : disables responses to broadcast probes
        -0               : set all WPA,WEP,open tags.
        -z <type>        : sets WPA1 tags.
        -Z <type>        : same as -z, but for WPA2
        -V <type>        : fake EAPOL 1=MD5 2=SHA1 3=auto
        -F <prefix>      : write all sent and received frames into pcap file
        -P               : respond to all probes, even when specifying ESSIDs
        -I <interval>    : sets the beacon interval value in ms
        -C <seconds>     : enables beaconing of probed ESSID values
        -n <hex>         : User specified ANonce when doing the 4-way handshake
        --bssid <MAC>    : BSSID to filter/use
        --bssids <file>  : read a list of BSSIDs out of that file
        --client <MAC>   : MAC of client to filter
        --clients <file> : read a list of MACs out of that file
        --essid <ESSID>  : specify a single ESSID (default: default)
        --essids <file>  : read a list of ESSIDs out of that file
        --help           : Displays this usage screen
    """

    command = "airbase-ng"
    requires_tempfile = False
    requires_tempdir = False
    requires_root = True



"""Airodump."""

import asyncio
import os
import xml
from contextlib import suppress

import dotmap
import xmltodict
from async_timeout import timeout

from .executor import ExecutorHelper
from .models import AccessPoint, Result


class AirodumpNg(ExecutorHelper):
    """Airodump-ng 1.6  - (C) 2006-2020 Thomas d'Otreppe
    https://www.aircrack-ng.org

    usage: airodump-ng <options> <interface>[,<interface>,...]

    Options:
        --ivs                 : Save only captured IVs
        --gpsd                : Use GPSd
        --write <prefix>      : Dump file prefix
        --beacons             : Record all beacons in dump file
        --update <secs>       : Display update delay in seconds
        --showack             : Prints ack/cts/rts statistics
        -h                    : Hides known stations for --showack
        -f <msecs>            : Time in ms between hopping channels
        --berlin <secs>       : Time before removing the AP/client
                                from the screen when no more packets
                                are received (Default: 120 seconds)
        -r  <file>            : Read packets from that file
        -T                    : While reading packets from a file,
                                simulate the arrival rate of them
                                as if they were "live".
        -x <msecs>            : Active Scanning Simulation
        --manufacturer        : Display manufacturer from IEEE OUI list
        --uptime              : Display AP Uptime from Beacon Timestamp
        --wps                 : Display WPS information (if any)
        --output-format <formats> : Output format. Possible values:
                                    pcap, ivs, csv, gps, kismet, netxml,
                                    logcsv
        --ignore-negative-one : Removes the message that says
                                fixed channel <interface>: -1
        --write-interval <seconds> : Output file(s) write interval in
                                     seconds
        --background <enable> : Override background detection.
        -n <int>              : Minimum AP packets recv'd before for
                                displaying it
        --encrypt <suite>     : Filter APs by cipher suite
        --netmask <netmask>   : Filter APs by mask
        --bssid <bssid>       : Filter APs by BSSID
        --essid <essid>       : Filter APs by ESSID
        --essid-regex <regex> : Filter APs by ESSID using a regular
                                expression
        -a                    : Filter unassociated clients
        --ht20                : Set channel to HT20 (802.11n)
        --ht40-               : Set channel to HT40- (802.11n)
        --ht40+               : Set channel to HT40+ (802.11n)
        --channel <channels>  : Capture on specific channels
        --band <abg>          : Band on which airodump-ng should hop
        -C <frequencies>      : Uses these frequencies in MHz to hop
        --cswitch <method>    : Set channel switching method
        -s                    : same as --cswitch
        --help                : Displays this usage screen
    """

    requires_tempfile = False
    requires_tempdir = True
    requires_root = True
    command = "airodump-ng"

    async def run(self, *args, **kwargs):
        """Run async, with prefix stablished as tempdir."""
        self.execn += 1

        kwargs = {
            "background": 1,
            "write": self.tempdir.name + "/" + self.uuid,
            "write-interval": 1,
            "output-format": "netxml,logcsv",
            **kwargs,
        }

        if "kismet" not in kwargs.get("output-format", ""):
            kwargs["output-format"] += ",netxml"

        return await super().run(*args, **kwargs)

    def get_file(self, file_format) -> str:
        """Return csv file, not kismet one.

        Arguments:

            file_format: File extension to retrieve (kismet.csv kismet.xml csv
                         log.csv or pcap)

        Returns: full filename
        """
        return f"{self.tempdir.name}/{self.uuid}-{self.execn:02}.{file_format}"

    @property
    async def results(self) -> list:
        """Return a list of currently detected access points.

        Returns: List of AccessPoint instances
        """
        file = self.get_file("kismet.netxml")
        try:
            # Wait for a sensible 3 seconds for netxml file to be generated and
            # process to be running
            async with timeout(3):
                while not os.path.exists(file):
                    await asyncio.sleep(1)

                while not self.proc:
                    # Check if airodump is running, otherwise wait more.
                    await asyncio.sleep(1)
        except asyncio.exceptions.TimeoutError as err:
            # No file had been generated or process hadn't started in 3
            # seconds.
            res = "Unknown"
            if self.proc:
                res = await self.proc.communicate()
            raise Exception("\n".join([a.decode() for a in res])) from err

        while self.running:
            # Avoid crashing on file creation
            with suppress(ValueError, xml.parsers.expat.ExpatError):
                xmla = xmltodict.parse(open(file).read())
                dotmap_data = dotmap.DotMap(xmla)
                results = dotmap_data["detection-run"]["wireless-network"]
                if results:
                    if isinstance(results, list):
                        return Result(
                            sorted(
                                [AccessPoint(ap) for ap in results],
                                key=lambda x: x.score,
                                reverse=True,
                            ))
                    else:
                        return Result([AccessPoint(results)])
                return Result([])

            await asyncio.sleep(1)
        return Result([])



import csv
import io
import re

from parse import parse
from rich.table import Table

DICTS = ("WLAN_", "JAZZTEL_", "MOVISTAR_")
APL_FMT = "{date}  Sending {attack} (code {num}) to {dest} -- BSSID: [{bssid}]"
MONITOR_RE = re.compile(
    r"\t\t\((\w+) (\w+) mode vif (\w+) for \[\w+\](\w+) on \[\w+\](\w+)\)")


class Result(list):

    @property
    def table(self):
        """Return a nicely formatted table with results."""
        table = Table(show_header=True,
                      header_style="bold magenta",
                      show_footer=False)
        if self:
            keys = self[0].asdict().keys()
            for key in keys:
                table.add_column(key.capitalize())

            for res in self:
                res = res.asdict()
                table.add_row(*[res[k] for k in keys])
        return table


class Interface:

    def __init__(self, data, monitor_data):
        self.data = data
        for data in monitor_data:
            if data["original_interface"] == self.data["interface"]:
                self.data[data["mode"]] = data

    @property
    def interface(self):
        return self.data["interface"]

    def __eq__(self, other):
        if isinstance(other, Interface):
            return other.interface == self.interface
        return self.interface == other

    @property
    def monitor(self):
        return self.data.get("monitor", {}).get("interface",
                                                self.data["interface"])

    def asdict(self):
        return self.data

    def __str__(self):
        return self.interface


class Interfaces(Result):

    def __init__(self, data):
        if data == [b"Run it as root"]:
            raise Exception("Pyrcrack must be run as root")
        pos = data.index(b"PHY	Interface	Driver		Chipset")
        ifaces_data = self.parse(b"\n".join(
            [a for a in data[pos:] if a and not a.startswith(b"\t\t")]))
        monitor_data = filter(lambda x: MONITOR_RE.match(x.decode()),
                              data[pos + len(ifaces_data):])

        def groups(data):
            return MONITOR_RE.match(data.decode()).groups()

        keys = ["driver", "mode", "status", "original_interface", "interface"]
        monitor_data = [dict(zip(keys, groups(a))) for a in monitor_data]
        self.extend([Interface(a, monitor_data) for a in ifaces_data])

    @staticmethod
    def parse(res):
        """Parse csv results."""
        with io.StringIO() as fileo:
            fileo.write(res.decode().strip().replace("\t\t", "\t"))
            fileo.seek(0)
            reader = csv.DictReader(fileo, dialect="excel-tab")
            return [{a.lower(): b for a, b in row.items()} for row in reader]


class Client:

    def __init__(self, data):
        self.data = data

    @property
    def bssid(self):
        return self.data["client-mac"]

    @property
    def packets(self):
        return getattr(self.data.packets, "total", 0)

    @property
    def dbm(self):
        return self.data["snr-info"].last_signal_dbm


class AccessPoint(dict):
    """Represents an access point.

    Stores internal data in "data" property
    """

    def __init__(self, data):
        """Initialize an access point.

        Arguments:

            data: Dot data structure from kismet xml file.
        """
        self.data = data
        self.update(self.asdict())

    def __repr__(self):
        return f"{self.essid} - ({self.bssid})"

    @property
    def airodump(self):
        return {"channel": self.channel, "bssid": self.bssid}

    @property
    def channel(self):
        return self.data.channel

    @property
    def encryption(self):
        return self.data.SSID.encryption

    @property
    def clients(self):
        """List of connected clients.

        Returns:

            List of Client instance.
        """
        if isinstance(self.data["wireless-client"], list):
            return [Client(d) for d in self.data["wireless-client"]]
        else:
            return [Client(self.data["wireless-client"])]

    @property
    def total_packets(self):
        return getattr(self.packets, "total", 0)

    def asdict(self):
        return {
            "essid": self.essid,
            "bssid": self.bssid,
            "packets": str(self.total_packets),
            "dbm": str(self.dbm),
            "score": str(self.score),
            "channel": str(self.channel),
            "encryption": "/".join(self.encryption),
        }

    @property
    def essid(self):
        """Essid."""
        return self.data.SSID.essid.get("#text", "")

    @property
    def bssid(self):
        """Mac (BSSID)"""
        return self.data.BSSID

    @property
    def score(self):
        """Score, used to sort networks.

        Score will take in account the total packets received, the dbm
        and if a ssid is susceptible to have dictionaries.
        """
        packet_score = int(self.total_packets)
        dbm_score = -int(self.dbm)
        dict_score = bool(any(self.essid.startswith(a) for a in DICTS))
        name_score = -1000 if not self.essid else 0
        enc_score = 1000 if "WEP" in self.encryption else 0
        return packet_score + dbm_score + dict_score + name_score + enc_score

    @property
    def packets(self):
        """Return list of packets."""
        return self.data.packets

    @property
    def dbm(self):
        """Return dbm info."""
        return self.data["snr-info"].last_signal_dbm

    def __lt__(self, other):
        """Compare with score."""
        return self.score


class AireplayResult(dict):
    """Single aiplay result line."""

    def __init__(self, *args, **kwargs):
        self.update(parse(APL_FMT, kwargs.pop("_data")).named)
        super().__init__(*args, **kwargs)

    def asdict(self):
        return self


class AireplayResults(Result):
    """Aireplay results."""

    def __init__(self, data):
        self.extend((AireplayResult(_data=a) for a in data.split("\n")
                     if "BSSID" in a))
